#info Kleinian groups
#info Original idea and code by Jos Leys (Dec 2016)
#info Some cosmetic modifications by Nat Merzouk Abdelaziz (A.k.a. Knighty)
#info liscence:  Ask Jos Leys ;-)

#include "ColorPalette.frag"

//#include "IBL-Raytracer.frag"
#define USE_EIFFIE_SHADOW
#define MULTI_SAMPLE_AO
//#define USE_IQ_CLOUDS

#define providesInit
#include "MathUtils.frag"
#include "DE-Kn2.frag"


#group Kleinian_test
uniform int Box_Iterations;  slider[1,50,100]
//Trace of the transformation a
uniform float KleinR;  slider[1.8,2.0,2.0]
uniform float KleinI;  slider[-0.2,0.0,0.2]
//angles cosines. if <1 it will be rounded to the nearest cos(PI/n). If >=1 it is free
uniform float AlphaX; slider[0.,0.,3.]
uniform float AlphaZ; slider[0.,0.,3.]
uniform float AlphaXZ; slider[0.,1.,5.]
//To really give a kleinian group they must be of the form cos(PI/n) if <1 and any value if >=1
//uniform float box_size_z;  slider[0.0,0.5,1.0]
//uniform float box_size_x;  slider[0.0,0.5,1.0]
//WIP: Help get better DE?
uniform int Final_Iterations;  slider[0,5,20]
//Want do show the balls
uniform bool ShowBalls; checkbox[true]
//4 generators or 3? apply y translation
uniform bool FourGen; checkbox[false]
//The DE is wilde. It needs some clamping
uniform float Clamp_y;  slider[0.01,1.0,10.0]
uniform float Clamp_DF;    slider[0.0,1.0,10.0]
uniform int ColorIterations;  slider[0,3,300]
uniform int ColorMode;  slider[1,1,6]
#group SphInv

//Want do do an inversion
uniform bool DoInversion; checkbox[false]
//Inversion center
uniform vec3 InvCenter; slider[(-1,-1,-1),(0,0,0),(1,1,1)]
//Inversion radius squared
uniform float InvRadius;  slider[0.01,1,2]
//Recenter
uniform vec3 ReCenter; slider[(-5,-5,-5),(0,0,0),(5,5,5)]

//initialization
float Tx=1., Tzz=1., Tzx=0.;

void init(){
	float ax=AlphaX, az=AlphaZ, axz=AlphaXZ;
	//Round angles to one of the form PI/n
#if 1
	if(ax<1.) ax = cos(PI/floor(PI/acos(ax)));
	if(az<1.) az = cos(PI/floor(PI/acos(az)));
	//if(axz<1.) axz = cos(PI/floor(PI/acos(axz)));
#endif
	//compute translations
	float lx = sqrt(2.*(1.+ ax)), lz = sqrt(2.*(1. + az));
	float ct = (lx*lx + lz*lz - 2.*(1.+ axz))/(2.*lx*lz);
	if(abs(ct)<1.) 
	{
		Tx = 0.5 * lx;
		Tzx = 0.5 * lz * ct;
		Tzz = 0.5 * lz * sqrt(1.-ct*ct);
	}
}
//---------------------
float dot2(vec3 z){ return dot(z,z);}

vec3 wrap(vec3 x, vec3 a){
	vec3 s = -0.5*a;
	x -= s; 
	return (x-a*floor(x/a)) + s;
}

vec2 wrap(vec2 x, vec2 a){
	vec2 s = -0.5*a;
	x -= s; 
	return (x-a*floor(x/a)) + s;
}

float wrap(float x, float a){
	float s = -0.5*a;
	x -= s; 
	return (x-a*floor(x/a)) + s;
}
float wrap2(float x, float a){
	return (x-a*floor(x/a));
}

vec3 correctz(vec3 z){
	float d2 = 1e30;
	float Tx2=2.*Tx, Tzx2=2.*Tzx, Tzz2=2.*Tzz;
	vec3 nn = vec3(0.);
	for(float nx=-1.; nx<=1.; nx+=1.)
		for(float ny=0.; ny<=0.; ny+=1.)
			for(float nz=-1.; nz<=1.; nz+=1.){
				float dd2=dot2(z-vec3(nx*Tx2+nz*Tzx2-ny*KleinI, ny*KleinR, nz*Tzz2));
				if(dd2<d2){
					d2=dd2;
					nn=vec3(nx,ny,nz);
				}
			}
	return z-vec3(nn.x*Tx2+nn.z*Tzx2-nn.y*KleinI, nn.y*KleinR, nn.z*Tzz2);
}
vec3 wrapit(vec3 z){
	z.x=z.x-Tzx/Tzz*z.z;
	if (FourGen){
		z.x=z.x+KleinI/KleinR*z.y;
		z.y = wrap2(z.y,  KleinR);
		z.xz = wrap(z.xz, 2. * vec2( Tx,  Tzz));
		z.x=z.x-KleinI/KleinR*z.y;
	} else
		z.xz = wrap(z.xz, 2. * vec2( Tx,  Tzz));
	z.x=z.x+Tzx/Tzz*z.z;
	return correctz(z);
}

void TransA(inout vec3 z, inout float DF, float a, float b){
	float iR = 1. / dot2(z);
	z *= iR;
	z.x = b + z.x; z.y = a - z.y; 
	DF *= iR;//max(1.,iR);
}

void TransAInv(inout vec3 z, inout float DF, float a, float b){
	//float iR = 1. / dot2(z + vec3(b,-a,0.));
	z.x -= b; z.y = a - z.y;
	float iR = 1. / dot2(z); 
	z *= iR;
	DF *= iR;//max(1.,iR);
}

float  JosKleinian(vec3 z)
{
	vec3 lz=z+vec3(1.), llz=z+vec3(-1.);
	float DE=1e10;
	float DF = 1.0;
	float a = KleinR, b = KleinI;
	float f = sign(b) ;     
	for (int i = 0; i < Box_Iterations ; i++) 
	{
		//if(z.y<0. || z.y>a) break;
		z = wrapit(z);
		//orbitTrap = min(orbitTrap, abs(vec4(z,dot(z,z))));//For colouring
#if 0
		//Use transformation a or its inverse by checking the position of z wrt the "separation line"
		if  (z.y >= a * (0.5 -  f * 0.25 * sign(z.x + b * 0.5)* (1. - exp( - 3. * abs(z.x + b * 0.5))))) //(z.y>0.5*a) //	
			TransAInv(z, DF, a, b);//Apply transformation a
		else
			TransA(z, DF, a, b);
#else
		//If above the separation line, rotate by 180 about (-b/2, a/2)
		if  (z.y >= a * (0.5 -  f * 0.15 * sign(z.x - b * 0.5)* (1. - exp( - 5.2 * abs(z.x - b * 0.5)))))
			z = vec3(b, a, 0.) - z;//
			//z.xy = vec2(-b, a) - z.xy;//
		TransA(z, DF, a, b);	
#endif
		orbitTrap = min(orbitTrap, abs(vec4(z,dot(z,z))));//For colouring
#if 1		
		//If the iterated points enters a 2-cycle , bail out.
		if(dot2(z-llz) < 1e-12) {
#if 0
			orbitTrap =vec4(1./float(i),0.,0.,0.);
#endif
			break;
		}
		//Store prvious iterates
		llz=lz; lz=z;
#endif
	}
	
	//WIP: Push the iterated point left or right depending on the sign of KleinI
	for (int i=0;i<Final_Iterations;i++){
		float y = ShowBalls ? min(z.y, a-z.y) : z.y;
		DE=min(DE,min(y,Clamp_y)/max(DF,Clamp_DF));
		TransA(z, DF, a, b);
	}
	float y = ShowBalls ? min(z.y, a-z.y) : z.y;
	DE=min(DE,min(y,Clamp_y)/max(DF,Clamp_DF));

	return DE;
}

float DE(vec3 p) {
	if(DoInversion){
		p=p-InvCenter-ReCenter;
		float r=length(p);
		float r2=r*r;
		p=(InvRadius * InvRadius/r2)*p+InvCenter;
		float de=JosKleinian(p);
		de=r2*de/(InvRadius * InvRadius+r*de);
		return de;
	}
	else return JosKleinian(p);
}

float Coloring(vec3 z) {
	float expsmooth=0.0;
	float r1 = 0.0, r2 = 0.0;
	
	vec3 lz=z+vec3(1.), llz=z+vec3(-1.);
	float DE=1e10;
	float DF = 1.0;
	float a = KleinR, b = KleinI;
	float f = sign(b) ;     
	for (int i=0; i<ColorIterations; i++) {
		z = wrapit(z);
		if(ColorMode==3 || ColorMode==5){r1 = r2;r2=dot(z,z);}
		if(ColorMode==4 || ColorMode==6){r1 = r2;r2=abs(vec4(z,dot(z,z)));}
//#if 0
		//Use transformation a or its inverse by checking the position of z wrt the "separation line"
		if  (z.y >= a * (0.5 -  f * 0.25 * sign(z.x + b * 0.5)* (1. - exp( - 3. * abs(z.x + b * 0.5))))) //(z.y>0.5*a) //	
			TransAInv(z, DF, a, b);//Apply transformation a
		else
			TransA(z, DF, a, b);
			
//#else
		//If above the separation line, rotate by 180 about (-b/2, a/2)
//		if  (z.y >= a * (0.5 -  f * 0.15 * sign(z.x - b * 0.5)* (1. - exp( - 5.2 * abs(z.x - b * 0.5)))))
//			z = vec3(b, a, 0.) - z;//
			//z.xy = vec2(-b, a) - z.xy;//
//		TransA(z, DF, a, b);	
		
//#endif
		//orbitTrap = min(orbitTrap, abs(vec4(z,dot(z,z))));//For colouring
		if(ColorMode==1 || ColorMode==5){r1 = r2;r2=dot(z,z);}
		if(ColorMode==2 || ColorMode==6){r1 = r2;r2=abs(vec4(z,dot(z,z)));}

		expsmooth+=exp(-1.0/abs(r1-r2));
	}	
	return expsmooth;
}

